#!/usr/bin/env python
#
# Copyright 2010 Luca Olivetti
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This files contains a couple of (yet incomplete) classes to manage 
# EPAKs and PAKS (files used to upgrade the firmware in LG tv)
#
# If run as a standalone program, will save the content of each
# PAK embedded in the EPAK (.epk) file
#

import sys
import struct

crc32_table=[]
CRC32_POLY = 0x04c11db7

for i in range(256):
  c = i << 24
  for j in range(8):
     if c & 0x80000000:
       c = (c << 1) ^ CRC32_POLY
     else:
       c = c << 1
  crc32_table.append(c & 0xFFFFFFFF)       

def calc_crc(data):
   crc = 0xffffffff;
   for c in data:
      indx=(crc >> 24) ^ ord(c)
      crc = ((crc << 8) ^ crc32_table[indx]) & 0xffffffff
   return crc    
   
  
class PAK:
   """Class to manage a single PAK embedded in a EPAK file"""
   def __init__(self,data):
     i=0
     self.type=data[i:i+4]
     i=i+4
     size=struct.unpack("I",data[i:i+4])[0]
     i=i+4
     self.modelname=data[i:i+64]
     i=i+64
     #version and date are bcd:
     #  date  YYYYMMDD
     #  version either MMmm (major, minor)
     #          or MMmmss (major, minor, subminor) ??    
     self.version, self.date, self.mode=struct.unpack("I I I",data[i:i+12])
     i=i+12
     self.reserved=data[i:i+44]
     i=i+44 #reserved
     self.data=data[i:i+size]
     i=i+size
     self.crc=struct.unpack("I",data[i:])[0]
     
   def save(self,filename=None):
      """Saves only the data to a file"""
      if not filename:
        filename=self.type
      f=open(filename,'w')
      f.write(self.data)
      f.close()
      
   def load(self,filename):
      """Loads the data from a file and recalculates the crc"""
      f=open(filename,'r')
      self.data=f.read()
      self.crc=calc_crc(self.data)
      
   def raw(self):
      """Returns the whole PAK to embed in an EPAK"""
      return self.type+struct.pack("I",len(self.data))+self.modelname+struct.pack("I I I",self.version,self.date,self.mode)+44*chr(0)+self.data+struct.pack("I",self.crc)
             

class EPAK:
  """Class to read/write an EPAK (.epk) file"""
  def __init__(self,filename=None):
    """Note that without data the epak is invalid"""
    self.paks = []
    if filename:
      self.read(filename)
  def read(self,filename):
    """Read and parse an epk file"""
    f=file(filename,'r')
    f.seek(0)
    b = f.read(12)
    a = struct.unpack("4s I I",b)
    self.type=a[0]
    if self.type!='epak':
      raise TypeError("This isn't an epak file")
    self.size=a[1]
    fileNum=a[2]
    self.paks =[]
    for i in range(fileNum):
      f.seek(12+i*8)
      b=f.read(8)
      a=struct.unpack("I I",b)
      #print "pak %d starts at %d" % (i,a[0])
      f.seek(a[0])
      self.paks.append(PAK(f.read(a[1])))
    f.seek(12+20*8)   
    b=f.read(4)
    self.version=struct.unpack("I",b)[0]    
    f.close()
    
  def pakindex(self,type):
    """Finds an embedded pak by name (type in LG terminology)"""
    for i in range(len(self.paks)):
      if self.paks[i].type==type:
         return i
    raise NameError("Pak %s not found" % type)
         
    
def main():
  """Main function that simply reads an .epk and write a file for each PAK"""
  if len(sys.argv)!=2:
    print "Usage: epak <epak-filename>"
    sys.exit(2)
  e=EPAK(sys.argv[1])
  print "EPAK %s, type: %s, version: %08x, contains %d paks" % (sys.argv[1], e.type, e.version, len(e.paks))
  for pak in e.paks:
    print "saving PAK %s, version: %08x, date: %08x, mode: %d, modelname: %s" % (pak.type, pak.version, pak.date, pak.mode, pak.modelname.partition(chr(0))[0])
    pak.save()


if __name__ == "__main__":
  main()     
   